@@ -1,5 +1,6 @@
-from flask import redirect, request, session
-from flask.ext.babelpkg import refresh
+from flask import redirect, session
+from flask_babel import refresh
+
 from ..baseviews import BaseView, expose
 
 
@@ -12,6 +13,5 @@ class LocaleView(BaseView):
     def index(self, locale):
         session['locale'] = locale
         refresh()
-        return redirect(self._get_redirect())
-
-
+        self.update_redirect()
+        return redirect(self.get_redirect())
