@@ -1,4 +1,4 @@
-from flask.ext.appbuilder.widgets import RenderTemplateWidget
+from flask_appbuilder.widgets import RenderTemplateWidget
 
 
 class ChartWidget(RenderTemplateWidget):
@@ -11,4 +11,3 @@ class DirectChartWidget(RenderTemplateWidget):
 
 class MultipleChartWidget(RenderTemplateWidget):
     template = 'appbuilder/general/widgets/multiple_chart.html'
-
